@@ -31,9 +31,11 @@
 #if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
 #include "ext2fs.h"
+#include "e2p.h"
 #else
 #include "ext2fs/ext2_fs.h"
 #include "ext2fs/ext2fs.h"
+#include "e2p/e2p.h"
 #endif
 
 #ifdef ENABLE_NLS
@@ -45,10 +47,12 @@
 #else
 #define N_(a) (a)
 #endif
-/* FIXME */
+#ifndef NLS_CAT_NAME
 #define NLS_CAT_NAME "e2fsprogs"
+#endif
+#ifndef LOCALEDIR
 #define LOCALEDIR "/usr/share/locale"
-/* FIXME */
+#endif
 #else
 #define _(a) (a)
 #define N_(a) a
@@ -123,6 +127,10 @@ extern errcode_t resize_fs(ext2_filsys fs, blk_t *new_size, int flags,
 					    int pass, unsigned long cur,
 					    unsigned long max));
 
+extern errcode_t adjust_fs_info(ext2_filsys fs, ext2_filsys old_fs, 
+				blk_t new_size);
+
+
 /* extent.c */
 extern errcode_t ext2fs_create_extent_table(ext2_extent *ret_extent,
 					    int size);
@@ -134,6 +142,10 @@ extern void ext2fs_extent_dump(ext2_extent extent, FILE *out);
 extern errcode_t ext2fs_iterate_extent(ext2_extent extent, __u32 *old_loc,
 				       __u32 *new_loc, int *size);
 
+/* online.c */
+extern errcode_t online_resize_fs(ext2_filsys fs, const char *mtpt, 
+				  blk_t *new_size, int flags);
+
 /* sim_progress.c */
 extern errcode_t ext2fs_progress_init(ext2_sim_progmeter *ret_prog,
 				      const char *label,
